# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Union

import torch
import wild_time_data
from torch.optim import AdamW, Optimizer
from torch.optim.lr_scheduler import CosineAnnealingLR, StepLR, _LRScheduler
from torchmetrics.classification import MulticlassAccuracy
from torchvision.transforms import transforms
from transformers import AutoTokenizer
from wild_time_data import default_transform

from renate.benchmark.datasets.nlp_datasets import HuggingFaceTextDataModule, MultiTextDataModule
from renate.benchmark.datasets.vision_datasets import (
    CDDBDataModule,
    CLEARDataModule,
    CORE50DataModule,
    DomainNetDataModule,
    TorchVisionDataModule,
)
from renate.benchmark.datasets.wild_time_data import WildTimeDataModule
from renate.benchmark.models import (
    MultiLayerPerceptron,
    LearningToPromptTransformer,
    ResNet18,
    ResNet18CIFAR,
    ResNet34,
    ResNet34CIFAR,
    ResNet50,
    ResNet50CIFAR,
    VisionTransformerB16,
    VisionTransformerB32,
    VisionTransformerCIFAR,
    VisionTransformerH14,
    VisionTransformerL16,
    VisionTransformerL32,
)
from renate.benchmark.models.transformer import HuggingFaceSequenceClassificationTransformer
from renate.benchmark.scenarios import (
    ClassIncrementalScenario,
    DataIncrementalScenario,
    FeatureSortingScenario,
    HueShiftScenario,
    IIDScenario,
    ImageRotationScenario,
    PermutationScenario,
    Scenario,
)
from renate.data.data_module import RenateDataModule
from renate.models import RenateModule
from renate.models.prediction_strategies import ICaRLClassificationStrategy

from renate.benchmark.models.spromptmodel import SPromptTransformer

models = {
    "MultiLayerPerceptron": MultiLayerPerceptron,
    "ResNet18CIFAR": ResNet18CIFAR,
    "ResNet34CIFAR": ResNet34CIFAR,
    "ResNet50CIFAR": ResNet50CIFAR,
    "ResNet18": ResNet18,
    "ResNet34": ResNet34,
    "ResNet50": ResNet50,
    "VisionTransformerCIFAR": VisionTransformerCIFAR,
    "VisionTransformerB16": VisionTransformerB16,
    "VisionTransformerB32": VisionTransformerB32,
    "VisionTransformerL16": VisionTransformerL16,
    "VisionTransformerL32": VisionTransformerL32,
    "VisionTransformerH14": VisionTransformerH14,
    "HuggingFaceTransformer": HuggingFaceSequenceClassificationTransformer,
    "LearningToPromptTransformer": LearningToPromptTransformer,
    "SPromptTransformer": SPromptTransformer,
}


def model_fn(
    num_outputs: int,
    model_state_url: Optional[str] = None,
    updater: Optional[str] = None,
    model_name: Optional[str] = None,
    num_inputs: Optional[int] = None,
    num_hidden_layers: Optional[int] = None,
    hidden_size: Optional[Tuple[int]] = None,
    dataset_name: Optional[str] = None,
    pretrained_model_name_or_path: Optional[str] = None,
    prompt_size: int = 10,
    clusters_per_task: int = 5,
    per_task_classifier: bool = True,
) -> RenateModule:
    """Returns a model instance."""
    if model_name not in models:
        raise ValueError(f"Unknown model `{model_name}`")
    model_class = models[model_name]
    model_kwargs = {"num_outputs": num_outputs}
    if updater == "Avalanche-iCaRL":
        model_kwargs["prediction_strategy"] = ICaRLClassificationStrategy()
    if model_name == "MultiLayerPerceptron":
        model_kwargs.update(
            {
                "num_inputs": num_inputs,
                "num_hidden_layers": num_hidden_layers,
                "hidden_size": hidden_size,
            }
        )
    elif model_name.startswith("ResNet") and dataset_name in ["FashionMNIST", "MNIST", "yearbook"]:
        model_kwargs["gray_scale"] = True
    elif model_name == "HuggingFaceTransformer":
        if updater == "Avalanche-iCaRL":
            raise ValueError("Transformers do not support iCaRL.")
        model_kwargs["pretrained_model_name_or_path"] = pretrained_model_name_or_path
    elif (updater is not None) and ("LearningToPrompt" in updater):
        if not model_name.startswith("LearningToPrompt"):
            raise ValueError(
                "L2P model updaters are designed to work only with "
                f"LearningToPromptTransformer, but model name specified is {model_name}."
            )
        model_kwargs["pretrained_model_name_or_path"] = pretrained_model_name_or_path
    elif (updater is not None) and ("SPeft" in updater):
        if not model_name.startswith("SPrompt"):
            raise ValueError(
                "SPrompt model updater is designed to work only with "
                f"SPromptTransformer, but model name specified is {model_name}."
            )
        model_kwargs["pretrained_model_name_or_path"] = pretrained_model_name_or_path
        model_kwargs["prompt_size"] = prompt_size
        model_kwargs["clusters_per_task"] = clusters_per_task
        model_kwargs["per_task_classifier"] = per_task_classifier
    if model_state_url is None:
        model = model_class(**model_kwargs)
    else:
        state_dict = torch.load(model_state_url)
        model = model_class.from_state_dict(state_dict)
    return model


def get_data_module(
    data_path: str,
    src_bucket: Optional[str],
    src_object_name: Optional[str],
    dataset_name: str,
    val_size: float,
    seed: int,
    pretrained_model_name_or_path: Optional[str],
    input_column: Optional[str],
    target_column: Optional[str],
) -> RenateDataModule:
    tokenizer = None
    if pretrained_model_name_or_path is not None and "vit" not in pretrained_model_name_or_path:
        tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path)
    if dataset_name in TorchVisionDataModule.dataset_dict:
        return TorchVisionDataModule(
            data_path, dataset_name=dataset_name, val_size=val_size, seed=seed
        )
    if dataset_name in ["CLEAR10", "CLEAR100"]:
        return CLEARDataModule(data_path, dataset_name=dataset_name, val_size=val_size, seed=seed)
    if dataset_name in wild_time_data.list_datasets():
        data_module_kwargs = {
            "data_path": data_path,
            "src_bucket": src_bucket,
            "src_object_name": src_object_name,
            "dataset_name": dataset_name,
            "val_size": val_size,
            "seed": seed,
        }
        if pretrained_model_name_or_path is not None:
            data_module_kwargs["tokenizer"] = tokenizer
        return WildTimeDataModule(**data_module_kwargs)
    if dataset_name == "DomainNet":
        return DomainNetDataModule(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
    if dataset_name == "MultiText":
        return MultiTextDataModule(
            data_path=data_path,
            tokenizer=tokenizer,
            data_id="ag_news",
            val_size=val_size,
            seed=seed,
        )
    if dataset_name.startswith("hfd-"):
        return HuggingFaceTextDataModule(
            data_path=data_path,
            dataset_name=dataset_name[4:],
            input_column=input_column,
            target_column=target_column,
            tokenizer=tokenizer,
            val_size=val_size,
            seed=seed,
        )
    if dataset_name == "CDDB":
        return CDDBDataModule(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
    if dataset_name == "Core50":
        return CORE50DataModule(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )

    raise ValueError(f"Unknown dataset `{dataset_name}`.")


def get_scenario(
    scenario_name: str,
    data_module: RenateDataModule,
    chunk_id: int,
    seed: int,
    num_tasks: Optional[int] = None,
    groupings: Optional[Tuple[Tuple[int]]] = None,
    degrees: Optional[List[int]] = None,
    input_dim: Optional[Union[List[int], Tuple[int], int]] = None,
    feature_idx: Optional[int] = None,
    randomness: Optional[float] = None,
    data_ids: Optional[Tuple[Union[int, str]]] = None,
) -> Scenario:
    """Function to create scenario based on name and arguments.

    Args:
        scenario_name: Name to identify scenario.
        data_module: The base data module.
        chunk_id: The data chunk to load in for the training or validation data.
        seed: A random seed to fix the created scenario.
        num_tasks: The total number of expected tasks for experimentation.
        groupings: Used for scenario `ClassIncrementalScenario` to partition datasets into chunks by
            class. Used by `DataIncrementalScenario` to group domains to chunks..
        degrees: Used for scenario `ImageRotationScenario`. Rotations applied for each chunk.
        input_dim: Used for scenario `PermutationScenario`. Input dimensionality.
        feature_idx: Used for scenario `SoftSortingScenario`. Index of feature to sort by.
        randomness: Used for all `_SortingScenario`. Randomness strength in [0, 1].
        data_ids: List of data_ids for the `DataIncrementalScenario`.

    Returns:
        An instance of the requested scenario.

    Raises:
        ValueError: If scenario name is unknown.
    """
    if scenario_name == "ClassIncrementalScenario":
        assert groupings is not None, "Provide `groupings` for the class-incremental scenario."
        return ClassIncrementalScenario(
            data_module=data_module,
            groupings=groupings,
            chunk_id=chunk_id,
        )
    if scenario_name == "IIDScenario":
        return IIDScenario(
            data_module=data_module, num_tasks=num_tasks, chunk_id=chunk_id, seed=seed
        )
    if scenario_name == "ImageRotationScenario":
        return ImageRotationScenario(
            data_module=data_module, degrees=degrees, chunk_id=chunk_id, seed=seed
        )
    if scenario_name == "PermutationScenario":
        return PermutationScenario(
            data_module=data_module,
            num_tasks=num_tasks,
            input_dim=input_dim,
            chunk_id=chunk_id,
            seed=seed,
        )
    if scenario_name == "FeatureSortingScenario":
        return FeatureSortingScenario(
            data_module=data_module,
            num_tasks=num_tasks,
            feature_idx=feature_idx,
            randomness=randomness,
            chunk_id=chunk_id,
            seed=seed,
        )
    if scenario_name == "HueShiftScenario":
        return HueShiftScenario(
            data_module=data_module,
            num_tasks=num_tasks,
            randomness=randomness,
            chunk_id=chunk_id,
            seed=seed,
        )
    if scenario_name == "DataIncrementalScenario":
        if data_ids is None and groupings is None:
            data_ids = [data_id for data_id in range(num_tasks)]
        return DataIncrementalScenario(
            data_module=data_module,
            chunk_id=chunk_id,
            data_ids=data_ids,
            groupings=groupings,
            seed=seed,
        )
    raise ValueError(f"Unknown scenario `{scenario_name}`.")


def loss_fn(updater: Optional[str] = None) -> torch.nn.Module:
    if updater.startswith("Avalanche-"):
        return torch.nn.CrossEntropyLoss()
    return torch.nn.CrossEntropyLoss(reduction="none")


def data_module_fn(
    data_path: str,
    chunk_id: int,
    seed: int,
    scenario_name: str,
    dataset_name: str,
    val_size: float = 0.0,
    num_tasks: Optional[int] = None,
    groupings: Optional[Tuple[Tuple[int]]] = None,
    degrees: Optional[Tuple[int]] = None,
    input_dim: Optional[Tuple[int]] = None,
    feature_idx: Optional[int] = None,
    randomness: Optional[float] = None,
    src_bucket: Optional[str] = None,
    src_object_name: Optional[str] = None,
    pretrained_model_name_or_path: Optional[str] = None,
    input_column: Optional[str] = None,
    target_column: Optional[str] = None,
    data_ids: Optional[List[Union[int, str]]] = None,
):
    data_module = get_data_module(
        data_path=data_path,
        src_bucket=src_bucket,
        src_object_name=src_object_name,
        dataset_name=dataset_name,
        val_size=val_size,
        seed=seed,
        pretrained_model_name_or_path=pretrained_model_name_or_path,
        input_column=input_column,
        target_column=target_column,
    )
    if dataset_name in wild_time_data.list_datasets() and num_tasks is None:
        num_tasks = len(wild_time_data.available_time_steps(dataset_name))
    return get_scenario(
        scenario_name=scenario_name,
        data_module=data_module,
        chunk_id=chunk_id,
        seed=seed,
        num_tasks=num_tasks,
        groupings=groupings,
        degrees=degrees,
        input_dim=input_dim,
        feature_idx=feature_idx,
        randomness=randomness,
        data_ids=data_ids,
    )


def _get_normalize_transform(dataset_name):
    if dataset_name in TorchVisionDataModule.dataset_stats:
        return transforms.Normalize(
            TorchVisionDataModule.dataset_stats[dataset_name]["mean"],
            TorchVisionDataModule.dataset_stats[dataset_name]["std"],
        )
    if dataset_name in ["CLEAR10", "CLEAR100"]:
        return transforms.Normalize(
            CLEARDataModule.dataset_stats[dataset_name]["mean"],
            CLEARDataModule.dataset_stats[dataset_name]["std"],
        )
    if dataset_name == "DomainNet":
        return transforms.Normalize(
            DomainNetDataModule.dataset_stats["all"]["mean"],
            DomainNetDataModule.dataset_stats["all"]["std"],
        )
    if dataset_name == "CDDB":
        return transforms.Normalize(
            CDDBDataModule.dataset_stats["CDDB"]["mean"],
            CDDBDataModule.dataset_stats["CDDB"]["std"],
        )
    if dataset_name == "Core50":
        return transforms.Normalize(
            CORE50DataModule.dataset_stats["Core50"]["mean"],
            CORE50DataModule.dataset_stats["Core50"]["std"],
        )


def train_transform(dataset_name: str, model_name: Optional[str] = None) -> Optional[Callable]:
    """Returns a transform function to be used in the training."""
    if dataset_name == "fmow":
        return default_transform(dataset_name)
    if dataset_name == "yearbook":
        if (model_name is not None) and (
            ("Transformer" in model_name) and (model_name != "VisionTransformerCIFAR")
        ):
            return transforms.Compose(
                [
                    transforms.ToPILImage(),
                    transforms.Resize(224),
                    transforms.RandomHorizontalFlip(),
                    default_transform(dataset_name),
                    transforms.Lambda(lambda x: x.repeat(3, 1, 1)),
                ]
            )
        else:
            return default_transform(dataset_name)
    if dataset_name in [
        "MNIST",
        "FashionMNIST",
        "MultiText",
    ] + wild_time_data.list_datasets() or dataset_name.startswith("hfd-"):
        return None
    if dataset_name in ["CIFAR10", "CIFAR100"]:
        if (model_name is not None) and (
            ("Transformer" in model_name) and (model_name != "VisionTransformerCIFAR")
        ):
            return transforms.Compose(
                [
                    transforms.RandomResizedCrop(
                        224, scale=(0.05, 1.0), ratio=(3.0 / 4.0, 4.0 / 3.0)
                    ),
                    transforms.RandomHorizontalFlip(p=0.5),
                ]
            )
        else:
            return transforms.Compose(
                [
                    transforms.RandomCrop(32, padding=4),
                    transforms.RandomHorizontalFlip(),
                    _get_normalize_transform(dataset_name),
                ]
            )
    if dataset_name in ["CLEAR10", "CLEAR100", "DomainNet"]:
        return transforms.Compose(
            [
                transforms.Resize(224),
                transforms.RandomCrop(224),
                transforms.ToTensor(),
                _get_normalize_transform(dataset_name),
            ]
        )
    if dataset_name == "CDDB":
        return transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ColorJitter(brightness=63 / 255),
                transforms.ToTensor(),
                _get_normalize_transform(dataset_name),
            ]
        )
    if dataset_name == "Core50":
        return transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                _get_normalize_transform(dataset_name),
            ]
        )
    raise ValueError(f"Unknown dataset `{dataset_name}`.")


def test_transform(
    dataset_name: str,
    model_name: Optional[str] = None,
) -> Optional[Callable]:
    """Returns a transform function to be used for validation or testing."""
    if dataset_name == "fmow":
        return default_transform(dataset_name)
    if dataset_name == "yearbook":
        if (model_name is not None) and (
            ("Transformer" in model_name) and (model_name != "VisionTransformerCIFAR")
        ):
            return transforms.Compose(
                [
                    transforms.ToPILImage(),
                    transforms.Resize(224),
                    default_transform(dataset_name),
                    transforms.Lambda(lambda x: x.repeat(3, 1, 1)),
                ]
            )
        else:
            return default_transform(dataset_name)
    if dataset_name in [
        "MNIST",
        "FashionMNIST",
        "MultiText",
    ] + wild_time_data.list_datasets() or dataset_name.startswith("hfd-"):
        return None
    if dataset_name in ["CIFAR10", "CIFAR100"]:
        if (model_name is not None) and (
            ("Transformer" in model_name) and (model_name != "VisionTransformerCIFAR")
        ):
            return transforms.Compose(
                [
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                ]
            )
        else:
            return _get_normalize_transform(dataset_name)
    if dataset_name in ["CLEAR10", "CLEAR100", "DomainNet"]:
        return transforms.Compose(
            [
                transforms.Resize(224),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                _get_normalize_transform(dataset_name),
            ]
        )
    if dataset_name == "CDDB":
        return transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                _get_normalize_transform(dataset_name),
            ]
        )
    if dataset_name == "Core50":
        return transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                _get_normalize_transform(dataset_name),
            ]
        )
    raise ValueError(f"Unknown dataset `{dataset_name}`.")


def lr_scheduler_fn(
    learning_rate_scheduler: Optional[str] = None,
    learning_rate_scheduler_step_size: int = 30,
    learning_rate_scheduler_gamma: float = 0.1,
    learning_rate_scheduler_interval: str = "epoch",
    learning_rate_scheduler_t_max: Optional[int] = None,
    learning_rate_scheduler_eta_min: float = 0,
) -> Tuple[Optional[Callable[[Optimizer], _LRScheduler]], str]:
    if learning_rate_scheduler == "StepLR":
        return (
            partial(
                StepLR,
                step_size=learning_rate_scheduler_step_size,
                gamma=learning_rate_scheduler_gamma,
            ),
            learning_rate_scheduler_interval,
        )
    elif learning_rate_scheduler == "CosineAnnealingLR":
        return (
            partial(
                CosineAnnealingLR,
                T_max=learning_rate_scheduler_t_max,
                eta_min=learning_rate_scheduler_eta_min,
            ),
            learning_rate_scheduler_interval,
        )
    elif learning_rate_scheduler is None:
        return None, learning_rate_scheduler_interval
    raise ValueError(f"Unknown scheduler `{learning_rate_scheduler}`.")


def metrics_fn(num_outputs: int) -> Dict:
    return {"accuracy": MulticlassAccuracy(num_classes=num_outputs, average="micro")}


def optimizer_fn(
    optimizer: str,
    learning_rate: float,
    weight_decay: float,
    momentum: float = 0.0,  # TODO: fix problem that occurs when removing this
) -> Callable:
    if optimizer == "AdamW":
        return partial(AdamW, lr=learning_rate, weight_decay=weight_decay)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import warnings
from typing import Optional

from pytorch_lightning.strategies import Strategy, StrategyRegistry

_SUPPORTED_STRATEGIES = [
    "ddp_find_unused_parameters_false",
    "ddp",
    "deepspeed",
    "deepspeed_stage_1",
    "deepspeed_stage_2",
    "deepspeed_stage_2_offload",
    "deepspeed_stage_3",
    "deepspeed_stage_3_offload",
    "deepspeed_stage_3_offload_nvme",
]
_UNSUPPORTED_STRATEGIES = [
    x for x in StrategyRegistry.available_strategies() if x not in _SUPPORTED_STRATEGIES
]


def create_strategy(devices: int = 1, strategy_name: Optional["str"] = None) -> Strategy:
    """Function returns a strategy object based on the number of devices queried
    and name of strategy"""

    devices = devices or 1
    if strategy_name in _UNSUPPORTED_STRATEGIES:
        raise ValueError(
            f"Current strategy: {strategy_name} is unsupported. Choose deepspeed variants or ddp."
        )
    if devices < 0:
        raise ValueError("Number of devices has to be at least 0.")

    elif devices == 1:
        # If one GPU, use standard training. Enabled by passing strategy=None
        # to pl.Trainer
        if strategy_name is not None:
            warnings.warn(f"With devices=1, strategy is ignored. But got {strategy_name}.")

        return None
    elif strategy_name in ["none", "None", None]:
        # Nothing is specified and devices > 1. Fall back to DDP
        return StrategyRegistry.get("ddp")

    elif "deepspeed" in strategy_name:
        strategy = StrategyRegistry.get(strategy_name)

        # TODO: This should be changed to instantiating Deepspeed and settting it in
        # the constructor. This works for nowbecause forcing PyTorch optimizer flag isn't used
        # anywhere by Deepspeed.
        strategy.config["zero_force_ds_cpu_optimizer"] = False
        return strategy

    else:
        # Something else happened. Fall back to whatever is happening.
        return StrategyRegistry.get(strategy_name)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.defaults import TASK_ID


@pytest.mark.slow
def test_renate_vision_transformer_init():
    pytest.helpers.get_renate_module_vision_transformer(sub_class="visiontransformercifar")
    pytest.helpers.get_renate_module_vision_transformer(sub_class="visiontransformerb16")
    pytest.helpers.get_renate_module_vision_transformer(sub_class="visiontransformerb32")
    pytest.helpers.get_renate_module_vision_transformer(sub_class="visiontransformerl16")
    pytest.helpers.get_renate_module_vision_transformer(sub_class="visiontransformerl32")
    pytest.helpers.get_renate_module_vision_transformer(sub_class="visiontransformerh14")


@pytest.mark.slow
@pytest.mark.parametrize(
    "sub_class,input_dim",
    [
        ["visiontransformercifar", (3, 32, 32)],
        ["visiontransformerb16", (3, 224, 224)],
        ["visiontransformerb32", (3, 224, 224)],
        ["visiontransformerl16", (3, 224, 224)],
        ["visiontransformerl32", (3, 224, 224)],
        ["visiontransformerh14", (3, 224, 224)],
    ],
)
def test_renate_vision_transformer_fwd(sub_class, input_dim):
    vit = pytest.helpers.get_renate_module_vision_transformer(sub_class=sub_class)

    x = torch.rand(5, *input_dim)
    y_hat = vit(x)

    assert y_hat.shape[0] == 5
    assert y_hat.shape[1] == 10


# The following numbers have been computed by
# for m in [VisionTransformerB16, VisionTransformerB32, VisionTransformerCIFAR,
#           VisionTransformerH14, VisionTransformerL16, VisionTransformerL32]:
#     print(len(list(m()._backbone.parameters())))
@pytest.mark.slow
@pytest.mark.parametrize(
    "sub_class, expected_num_params",
    [
        ["visiontransformercifar", 54],
        ["visiontransformerb16", 198],
        ["visiontransformerb32", 198],
        ["visiontransformerl16", 390],
        ["visiontransformerl32", 390],
        ["visiontransformerh14", 518],
    ],
)
def test_renate_vision_transformer_get_params(sub_class, expected_num_params):
    vit = pytest.helpers.get_renate_module_vision_transformer(sub_class=sub_class)
    vit.add_task_params(TASK_ID)

    first_task_params = vit.get_params(TASK_ID)

    vit.add_task_params("second_task")

    second_task_params = vit.get_params("second_task")
    assert len(first_task_params) == len(second_task_params)
    assert len(first_task_params) == expected_num_params + 2
    assert len(second_task_params) == expected_num_params + 2
    assert len(list(vit.parameters())) == expected_num_params + 4
    for i in range(len(first_task_params) - 2):
        # -2 because the last two parameters are weight and bias of a task specific linear layer
        assert torch.equal(first_task_params[i], second_task_params[i])

    for i in range(len(first_task_params) - 2, len(first_task_params)):
        # -2 because the last two parameters are weight and bias of a task specific linear layer
        assert not torch.equal(first_task_params[i], second_task_params[i])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, Set

import torch

from renate.models.layers import ContinualNorm
from renate.types import NestedTensors
from renate.utils.deepspeed import convert_to_tensor, recover_object_from_tensor


class RenateModule(torch.nn.Module, ABC):
    """A class for torch models with some additional functionality for continual learning.

    ``RenateModule`` derives from ``torch.nn.Module`` and provides some additional functionality
    relevant to continual learning. In particular, this concerns saving and reloading the model
    when model hyperparameters (which might affect the architecture) change during hyperparameter
    optimization. There is also functionality to retrieve internal-layer representations for use
    in replay-based CL methods.

    When implementing a subclass of ``RenateModule``, make sure to call the base class' constructor
    and provide your model's constructor arguments. Besides that, you can define a
    ``RenateModule`` just like ``torch.nn.Module``.

    Example::

        class MyMNISTMLP(RenateModule):

        def __init__(self, num_hidden: int):
            super().__init__(
                constructor_arguments={"num_hidden": num_hidden}
                loss_fn=torch.nn.CrossEntropyLoss()
            )
            self._fc1 = torch.nn.Linear(28*28, num_hidden)
            self._fc2 = torch.nn.Linear(num_hidden, 10)

        def forward(self, x: torch.Tensor) -> torch.Tensor:
            x = self._fc1(x)
            x = torch.nn.functional.relu(x)
            return self._fc2(x)

    The state of a ``RenateModule`` can be retrieved via the ``RenateModule.state_dict()`` method,
    just as in ``torch.nn.Module``. When reloading a ``RenateModule`` from a stored state dict, use
    ``RenateModule.from_state_dict``. It wil automatically recover the hyperparameters and
    reinstantiate your model accordingly.

    Note: Some methods of ``RenateModule`` accept an optional ``task_id`` argument. This is in
    anticipation of future methods for continual learning scenarios where task identifiers are
    provided. It is currently not used.

    Args:
        constructor_arguments: Arguments needed to instantiate the model.
    """

    def __init__(self, constructor_arguments: dict):
        super(RenateModule, self).__init__()
        self._constructor_arguments = copy.deepcopy(constructor_arguments)
        self._tasks_params_ids: Set[str] = set()
        self._intermediate_representation_cache: List[torch.Tensor] = []
        self._hooks: List[Callable] = []

    @classmethod
    def from_state_dict(cls, state_dict):
        """Load the model from a state dict.

        Args:
            state_dict: The state dict of the model. This method works under the assumption that
                this has been created by `RenateModule.state_dict()`.
        """
        extra_state = recover_object_from_tensor(state_dict["_extra_state"])
        constructor_arguments = extra_state["constructor_arguments"]
        model = cls(**constructor_arguments)
        for task in extra_state["tasks_params_ids"]:
            model.add_task_params(task)
        # TODO: See https://github.com/awslabs/Renate/issues/236.
        # There are changes to the `class_means` or `componenets` of a model
        # that are not loaded, and should probably not be stored.
        model.load_state_dict(state_dict, strict=False)
        return model

    def get_extra_state(self, encode: bool = True) -> Any:
        """Get the constructor_arguments, and task ids necessary to reconstruct the model."""
        extra_state = {
            "constructor_arguments": self._constructor_arguments,
            "tasks_params_ids": self._tasks_params_ids,
        }
        return convert_to_tensor(extra_state) if encode else extra_state

    def set_extra_state(self, state: Any, decode: bool = True):
        """Extract the content of the ``_extra_state`` and set the related values in the module."""
        extra_state = recover_object_from_tensor(state) if decode else state
        self._constructor_arguments = extra_state["constructor_arguments"]
        self._tasks_params_ids = extra_state["tasks_params_ids"]

    @abstractmethod
    def forward(self, x: NestedTensors, task_id: Optional[str] = None) -> torch.Tensor:
        """Performs a forward pass on the inputs and returns the predictions.

        This method accepts a task ID, which may be provided by some continual learning scenarios.
        As an example, the task id may be used to switch between multiple output heads.

        Args:
            x: Input(s) to the model. Can be a single tensor, a tuple of tensor, or a dictionary
                mapping strings to tensors.
            task_id: The identifier of the task for which predictions are made.
        Returns:
            The model's predictions.
        """
        pass

    def get_params(self, task_id: Optional[str] = None) -> List[torch.nn.Parameter]:
        """User-facing function which returns the list of parameters.

        If a ``task_id`` is given, this should return only parameters used for the specific task.

        Args:
            task_id: The task id for which we want to retrieve parameters.
        """
        return list(self.parameters())

    def _add_task_params(self, task_id: str) -> None:
        """Adds new parameters, associated to a specific task, to the model.

        The method should not modify modules created in previous calls, beyond the ones defined
        in ``self._add_task_params()``. The order of the calls is not guaranteed when the model
        is loaded after being saved.

        Args:
            task_id: The task id for which the new parameters are added.
        """
        pass

    def add_task_params(self, task_id: Optional[str] = None) -> None:
        """Adds new parameters, associated to a specific task, to the model.

        This function should not be overwritten; use ``_add_task_params`` instead.

        Args:
            task_id: The task id for which the new parameters are added.
        """
        if task_id in self._tasks_params_ids:
            return
        self._add_task_params(task_id)
        self._tasks_params_ids.add(task_id)

    def get_logits(self, x: NestedTensors, task_id: Optional[str] = None) -> torch.Tensor:
        """Returns the logits for a given pair of input and task id.

        By default, this method returns the output of the forward pass. This may be overwritten
        with custom behavior, if necessary.

        Args:
            x: Input(s) to the model. Can be a single tensor, a tuple of tensor, or a dictionary
                mapping strings to tensors.
            task_id: The task id.
        """
        return self.forward(x, task_id)

    def get_intermediate_representation(self) -> List[torch.Tensor]:
        """Returns the cached intermediate representation."""
        return self._intermediate_representation_cache

    def replace_batch_norm_with_continual_norm(self, num_groups: int = 32) -> None:
        """Replaces every occurence of batch normalization with continual normalization.

        Pham, Q., Liu, C., & Hoi, S. (2022). Continual normalization: Rethinking batch
        normalization for online continual learning. arXiv preprint arXiv:2203.16102.

        Args:
            num_groups: Number of groups when considering the group normalization in continual
                normalization.
        """

        def _replace(module):
            for name, child in module.named_children():
                if not list(module.children()):
                    _replace(child)
                if isinstance(
                    child, (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d, torch.nn.BatchNorm3d)
                ):
                    setattr(
                        module,
                        name,
                        ContinualNorm(
                            num_features=child.num_features,
                            eps=child.eps,
                            momentum=child.momentum,
                            affine=child.affine,
                            track_running_stats=child.track_running_stats,
                            num_groups=num_groups,
                        ),
                    )

        _replace(self)

    def _intermediate_representation_caching_hook(self) -> Callable:
        """Hook to cache intermediate representations during training."""

        def hook(m: torch.nn.Module, _, output: torch.Tensor) -> None:
            if m.training:
                self._intermediate_representation_cache.append(output)

        return hook

    def register_intermediate_representation_caching_hook(self, module: torch.nn.Module) -> None:
        """Add a hook to cache intermediate representations during training.

        Store the reference to the hook to enable its removal.

        Args:
            module: The module to be hooked.
        """
        hook = module.register_forward_hook(self._intermediate_representation_caching_hook())
        self._hooks.append(hook)

    def deregister_hooks(self) -> None:
        """Remove all the hooks that were registered."""
        for hook in self._hooks:
            hook.remove()
        self._hooks = []
        self.reset_intermediate_representation_cache()

    def reset_intermediate_representation_cache(self) -> None:
        """Resets the intermediate representation cache."""
        self._intermediate_representation_cache = []


class RenateWrapper(RenateModule):
    """A simple wrapper around a torch model.

    If you are using a torch model with fixed hyperparameters, you can use this wrapper to expose
    it as a ``RenateModule``. In this case, do _not_ use the ``from_state_dict`` method but
    reinstantiate the model, wrap it, and call ``load_state_dict``. If a tuple or a dictionary of
    tensors is passed to the `RenateWrapper`'s forward function, it is unpacked before passing it
    to the torch model's forward function.

    Example::

        my_torch_model = torch.nn.Linear(28*28, 10)  # Instantiate your torch model.
        model = RenateWrapper(my_torch_model)
        state_dict = torch.load("my_state_dict.pt")
        model.load_state_dict(state_dict)

    Args:
        model: The torch model to be wrapped.
    """

    def __init__(self, model: torch.nn.Module) -> None:
        super().__init__(constructor_arguments={})
        self._model = model

    def forward(self, x: NestedTensors, task_id: Optional[str] = None) -> torch.Tensor:
        if isinstance(x, torch.Tensor):
            outputs = self._model(x)
        elif isinstance(x, tuple):
            outputs = self._model(*x)
        elif isinstance(x, dict):
            outputs = self._model(**x)
        else:
            raise TypeError(f"Expected tensor or tuple/dict of tensors; found {type(x)}.")

        if isinstance(outputs, tuple) and len(outputs) == 1:
            return outputs[0]
        else:
            return outputs

    @classmethod
    def from_state_dict(cls, state_dict):
        raise NotImplementedError(
            "RenateWrapper does not support `from_state_dict`. Instantiate the object using the "
            "standard constructor, then call `load_state_dict`."
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from renate.training import run_training_job

configuration = {
    "optimizer": "SGD",
    "momentum": 0.0,
    "weight_decay": 1e-2,
    "learning_rate": 0.05,
    "batch_size": 64,
    "batch_memory_frac": 0.5,
    "max_epochs": 50,
    "memory_size": 500,
}

if __name__ == "__main__":
    run_training_job(
        config_space=configuration,
        mode="max",
        metric="val_accuracy",
        updater="ER",
        max_epochs=50,
        chunk_id=0,
        config_file="renate_config.py",
        output_state_url="./output_folder/",  # this is where the model will be stored
        backend="local",  # the training job will run on the local machine
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from collections import Counter
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import torch
from avalanche.benchmarks import dataset_benchmark
from avalanche.core import BasePlugin
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from renate.data.datasets import _TransformedDataset
from renate.memory import DataBuffer


class BaseAvalancheDataset(Dataset):
    """Base class for all datasets consumable by Avalanche updaters."""

    def __init__(
        self,
        targets: List[int],
        collate_fn: Optional[Callable] = None,
    ):
        self._targets = targets
        self.targets = torch.tensor(targets, dtype=torch.long)
        if collate_fn is not None:
            self.collate_fn = collate_fn

    def __len__(self) -> int:
        return len(self._targets)


class AvalancheDataset(BaseAvalancheDataset):
    """A wrapper around a Dataset consumable by Avalanche updaters."""

    def __init__(
        self,
        dataset: Union[Dataset, DataBuffer],
        targets: List[int],
        collate_fn: Optional[Callable] = None,
    ):
        super().__init__(targets, collate_fn)
        self._dataset = dataset

    def __getitem__(self, idx) -> Tuple[Tensor, int]:
        return self._dataset[idx][0], self._targets[idx]


class AvalancheDatasetForBuffer(BaseAvalancheDataset):
    """A wrapper around a DataBuffer consumable by Avalanche updaters."""

    def __init__(
        self, buffer: DataBuffer, targets: List[int], collate_fn: Optional[Callable] = None
    ):
        super().__init__(targets, collate_fn)
        self._indices = buffer._indices
        self._datasets = buffer._datasets

    def __getitem__(self, idx) -> Tuple[Tensor, int]:
        i, j = self._indices[idx]
        return self._datasets[i][j][0], self._targets[idx]


def to_avalanche_dataset(
    dataset: Union[Dataset, DataBuffer], collate_fn: Optional[Callable] = None
) -> BaseAvalancheDataset:
    """Converts a DataBuffer or Dataset into an Avalanche-compatible Dataset."""
    y_data = []
    is_buffer = isinstance(dataset, DataBuffer)
    for i in range(len(dataset)):
        if is_buffer:
            (_, y), _ = dataset[i]
        else:
            _, y = dataset[i]
        if not isinstance(y, int):
            y = y.item()
        y_data.append(y)
    if is_buffer:
        return AvalancheDatasetForBuffer(dataset, y_data, collate_fn)
    return AvalancheDataset(dataset, y_data, collate_fn)


class AvalancheBenchmarkWrapper:
    def __init__(
        self,
        train_dataset,
        val_dataset,
        train_transform,
        train_target_transform,
        test_transform,
        test_target_transform,
    ):
        self._n_classes_per_exp = None
        self._classes_order = None
        self._n_classes = 0
        self._train_dataset = train_dataset
        self._test_transform = test_transform
        self._train_target_transform = train_target_transform
        self._benchmark = dataset_benchmark(
            [train_dataset],
            [val_dataset],
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            eval_transform=test_transform,
            eval_target_transform=test_target_transform,
        )
        self.train_stream = self._benchmark.train_stream
        self.test_stream = self._benchmark.test_stream

    def update_benchmark_properties(self):
        dataset = _TransformedDataset(
            dataset=self._train_dataset,
            transform=self._test_transform,
            target_transform=self._train_target_transform,
        )
        dataloader = DataLoader(dataset)
        unique_classes = set()
        for batch in dataloader:
            unique_classes.add(batch[1].item())
        if self._n_classes_per_exp is None:
            self._n_classes_per_exp = [len(unique_classes)]
            self._classes_order = list(sorted(unique_classes))
        else:
            self._n_classes_per_exp.append(len(unique_classes))
            self._classes_order += list(sorted(unique_classes))
        self._n_classes = sum(self._classes_order)
        self._benchmark.n_classes_per_exp = self._n_classes_per_exp
        self._benchmark.classes_order = self._classes_order

    def state_dict(self) -> Dict[str, Any]:
        """Returns the state of the benchmark."""
        state_dict = {
            "n_classes_per_exp": self._n_classes_per_exp,
            "classes_order": self._classes_order,
        }
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Restores the state of the benchmark."""
        self._n_classes_per_exp = state_dict["n_classes_per_exp"]
        self._classes_order = state_dict["classes_order"]
        self._benchmark.n_classes_per_exp = self._n_classes_per_exp
        self._benchmark.classes_order = self._classes_order


def replace_plugin(plugin: Optional[BasePlugin], plugins: List[BasePlugin]) -> List[BasePlugin]:
    """Replaces a plugin if already exists and appends otherwise.

    Args:
        plugin: New plugin that replaces existing one.
        plugins: List of current plugins.
    Returns:
        Reference to ``plugins``.
    """
    idx = _plugin_index(type(plugin), plugins)
    if idx >= 0:
        plugins[idx] = plugin
    else:
        plugins.append(plugin)
    return plugins


def remove_plugin(plugin_class: Type[BasePlugin], plugins: List[BasePlugin]) -> List[BasePlugin]:
    """Removes a plugin by class if exists.

    Args:
        plugin_class: Remove a plugin of this class if exists.
        plugins: List of current plugins.
    Returns:
        Reference to ``plugins``.
    """
    idx = _plugin_index(plugin_class, plugins)
    if idx >= 0:
        del plugins[idx]
    return plugins


def plugin_by_class(
    plugin_class: Type[BasePlugin], plugins: List[BasePlugin]
) -> Optional[BasePlugin]:
    """Returns plugin with respective class from a list of plugins.

    Args:
        plugin_class: Class type of interest in ``plugins``.
        plugins: List of plugins we search for an object of type ``plugin_class``.
    Returns:
        ``None`` if class does not exist, otherwise the respective object.
    """
    idx = _plugin_index(plugin_class, plugins)
    if idx >= 0:
        return plugins[idx]
    return None


def _plugin_index(plugin_class: Type[BasePlugin], plugins: List[BasePlugin]) -> int:
    """Returns index at which a plugin of that type is located in the list.

    Returns:
        Returns location of plugin and ``-1`` if it does not exist.
    """
    plugins_types = [type(p) for p in plugins]
    if max(Counter(plugins_types).values()) > 1:
        raise ValueError("Multiple occurrences of same type in `plugins` are not supported.")
    if plugin_class in plugins_types:
        return plugins_types.index(plugin_class)
    return -1



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
import os
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import torch
import torch.nn as nn
import torchmetrics
from pytorch_lightning import LightningModule
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch import Tensor
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.data.datasets import _TransformedDataset
from renate.memory import DataBuffer, InfiniteBuffer, ReservoirBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.utils.misc import maybe_populate_mask_and_ignore_logits
from renate.utils.pytorch import get_generator, unique_classes


class RenateLightningModule(LightningModule, abc.ABC):
    """Base class for LightningModules, which implement metric logging and basic training logic.

    The `RenateLightningModule` is a `LightningModule`, but provides additional hook functions
    called by `ModelUpdater`. These hooks are:

    - `on_model_update_start`, which is called in the beginning of a
       model update. We expect this to return train and (optionally) validation
       data loader(s).
    - `on_model_update_end`, which is called in the end of a model update.

    Args:
        model: The model to be trained.
        optimizer: Partial optimizer used to create an optimizer by passing the model parameters.
        learning_rate_scheduler: Partial object of learning rate scheduler that will be created by
            passing the optimizer.
        learning_rate_scheduler_interval: When to update the learning rate scheduler.
            Options: `epoch` and `step`.
        batch_size: Training batch size.
        logged_metrics: Metrics logged additional to the default ones.
        seed: See :func:`renate.models.utils.get_generator`.
        mask_unused_classes: Flag to use if logits corresponding to unused classes are to be ignored
            in the loss computation. Possibly useful for class incremental learning.
    """

    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        learning_rate_scheduler: Optional[Optional[Callable[[Optimizer], _LRScheduler]]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        seed: int = defaults.SEED,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ) -> None:
        super().__init__()
        self._model = model
        self._loss_fn = loss_fn
        self._optimizer = optimizer
        self._learning_rate_scheduler = learning_rate_scheduler
        self._learning_rate_scheduler_interval = learning_rate_scheduler_interval
        self._batch_size = batch_size
        self._seed = seed
        self._mask_unused_classes = mask_unused_classes

        self._class_mask = None
        self._classes_in_current_task = None

        self._task_id: str = defaults.TASK_ID
        self._train_dataset: Optional[Dataset] = None
        self._val_dataset: Optional[Dataset] = None
        self.val_enabled = False
        self._train_collate_fn: Optional[Callable] = None
        self._val_collate_fn: Optional[Callable] = None

        self._create_metrics_collections(logged_metrics)
        self._rng = get_generator(self._seed)
        self.save_hyperparameters(ignore=self._ignored_hyperparameters())

    def _ignored_hyperparameters(self):
        """Hyperparameters to be ignored in the ``save_hyperparameters`` call."""
        return [
            "model",
            "loss_fn",
            "optimizer",
            "learning_rate_scheduler",
            "logged_metrics",
        ]

    def _create_metrics_collections(
        self, logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None
    ) -> None:
        """Creates all logged metrics."""
        if logged_metrics is None:
            logged_metrics = {}
        metrics = torchmetrics.MetricCollection(logged_metrics)
        train_metrics = metrics.clone(prefix="train_")
        val_metrics = metrics.clone(prefix="val_")

        train_losses = nn.ModuleDict(
            {
                "base_loss": torchmetrics.MeanMetric(),
                "loss": torchmetrics.MeanMetric(),
            }
        )
        val_losses = nn.ModuleDict({"loss": torchmetrics.MeanMetric()})

        self._metric_collections = nn.ModuleDict(
            {
                "train_metrics": train_metrics,
                "val_metrics": val_metrics,
            }
        )
        self._loss_collections = nn.ModuleDict(
            {
                "train_losses": train_losses,
                "val_losses": val_losses,
            }
        )

    def is_logged_metric(self, metric_name: str) -> bool:
        """Returns `True` if there is a metric with name `metric_name`."""
        if metric_name is None:
            return True
        logged_metrics = list()
        for prefix in ["train", "val"]:
            for collection, collection_name in zip(
                [self._metric_collections, self._loss_collections], ["metrics", "losses"]
            ):
                collection_key = f"{prefix}_{collection_name}"
                if collection_key in collection:
                    logged_metrics += [
                        f"{prefix}_{logged_metric_name}"
                        for logged_metric_name in collection[collection_key]
                    ]
        return metric_name in logged_metrics

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        self._train_dataset = train_dataset
        self._val_dataset = val_dataset
        self.val_enabled = val_dataset is not None and len(val_dataset) > 0
        self._train_collate_fn = train_dataset_collate_fn
        self._val_collate_fn = val_dataset_collate_fn
        self._task_id = task_id
        self._model.add_task_params(task_id=self._task_id)
        if self._mask_unused_classes:
            # The first forward prop will populate the _class_mask with the following
            # unique classes
            self._classes_in_current_task = unique_classes(self._train_dataset)

    def train_dataloader(self) -> DataLoader:
        """Returns the dataloader for training the model."""
        return DataLoader(
            self._train_dataset,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
            collate_fn=self._train_collate_fn,
        )

    def val_dataloader(self) -> Optional[DataLoader]:
        if self._val_dataset is not None:
            return DataLoader(
                self._val_dataset,
                batch_size=self._batch_size,
                shuffle=False,
                generator=self._rng,
                pin_memory=True,
                collate_fn=self._val_collate_fn,
            )

    def on_model_update_end(self) -> None:
        """Called right before a model update terminates."""
        pass

    def forward(self, inputs: NestedTensors, task_id: Optional[str] = None) -> torch.Tensor:
        """Forward pass of the model."""
        if task_id is None:
            task_id = self._task_id
        return self._model(inputs, task_id=task_id)

    def training_step_unpack_batch(self, batch: Tuple[Any, Any]) -> Tuple[Any, Any]:
        inputs, targets = batch
        return inputs, targets

    def training_step(
        self, batch: Tuple[NestedTensors, torch.Tensor], batch_idx: int
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        inputs, targets = self.training_step_unpack_batch(batch)
        outputs = self(inputs)
        outputs, self._class_mask = maybe_populate_mask_and_ignore_logits(
            self._mask_unused_classes, self._class_mask, self._classes_in_current_task, outputs
        )
        intermediate_representation = self._model.get_intermediate_representation()
        self._model.reset_intermediate_representation_cache()
        loss = self._loss_fn(outputs, targets).mean()
        self._update_metrics(outputs, targets, "train")
        self._loss_collections["train_losses"]["base_loss"](loss)
        return {
            "loss": loss,
            "outputs": outputs,
            "intermediate_representation": intermediate_representation,
        }

    def training_step_end(self, step_output: STEP_OUTPUT) -> STEP_OUTPUT:
        """PyTorch Lightning function to perform after the training step."""
        super().training_step_end(step_output)
        self._loss_collections["train_losses"]["loss"](step_output["loss"])
        return step_output

    def training_epoch_end(self, outputs: List[Union[Tensor, Dict[str, Any]]]) -> None:
        """PyTorch Lightning function to run at the end of training epoch."""
        super().training_epoch_end(outputs)
        if not self.val_enabled:
            self._log_metrics()

    def validation_step_unpack_batch(self, batch: Tuple[Tuple[Any, Any], Any]) -> Tuple[Any, Any]:
        (inputs, targets), _ = batch
        return inputs, targets

    def validation_step(self, batch: Tuple[NestedTensors, torch.Tensor], batch_idx: int) -> None:
        """PyTorch Lightning function to estimate validation metrics."""
        inputs, targets = self.validation_step_unpack_batch(batch)
        outputs = self(inputs)
        loss = self._loss_fn(outputs, targets)
        self._update_metrics(outputs, targets, "val")
        self._loss_collections["val_losses"]["loss"](loss)

    def validation_epoch_end(self, outputs: List[Union[Tensor, Dict[str, Any]]]) -> None:
        """PyTorch Lightning function to run at the end of validation epoch."""
        super().validation_epoch_end(outputs)
        self._log_metrics()

    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Tuple[List[Optimizer], List[Dict[str, Any]]]]:
        """PyTorch Lightning function to create optimizers and learning rate schedulers."""
        optimizer = self._optimizer(self._model.get_params(self._task_id))
        if self._learning_rate_scheduler is None:
            return optimizer
        lr_scheduler_config = {
            "scheduler": self._learning_rate_scheduler(optimizer),
            "interval": self._learning_rate_scheduler_interval,
        }
        return [optimizer], [lr_scheduler_config]

    def _update_metrics(
        self,
        outputs: torch.Tensor,
        y: torch.Tensor,
        prefix: Literal["train", "val"],
    ) -> None:
        """Shared logic for updating metrics."""
        self._metric_collections[f"{prefix}_metrics"](outputs, y)

    def _log_metrics(
        self,
    ) -> None:
        """Shared logic for logging metrics, including the loss."""
        if self.trainer.sanity_checking:
            return
        prefixes = ["train", "val"] if self.val_enabled else ["train"]
        for prefix in prefixes:
            self.log_dict(
                self._metric_collections[f"{prefix}_metrics"].compute(),
                on_step=False,
                on_epoch=True,
                logger=True,
                sync_dist=True,
            )
            self._metric_collections[f"{prefix}_metrics"].reset()

            for loss_name, loss in self._loss_collections[f"{prefix}_losses"].items():
                self.log(
                    f"{prefix}_{loss_name}",
                    loss.compute(),
                    on_step=False,
                    on_epoch=True,
                    logger=True,
                    sync_dist=True,
                )
                loss.reset()


class Learner(RenateLightningModule, abc.ABC):
    """Base class for Learners, which encapsulate the core CL methodologies.

    The `Learner` is a `LightningModule`, but provides additional hook functions
    called by `ModelUpdater`. These hooks are:

    - `Learner.on_model_update_start`, which is called in the beginning of a
       model update. We expect this to return train and (optionally) validation
       data loader(s).
    - `Learner.on_model_update_end`, which is called in the end of a model update.

    This base class implements a basic training loop without any mechanism to
    counteract forgetting.

    Args:
        model: The model to be trained.
        optimizer: Partial optimizer used to create an optimizer by passing the model parameters.
        learning_rate_scheduler: Partial object of learning rate scheduler that will be created by
            passing the optimizer.
        learning_rate_scheduler_interval: When to update the learning rate scheduler.
            Options: `epoch` and `step`.
        batch_size: Training batch size.
        train_transform: The transformation applied during training.
        train_target_transform: The target transformation applied during testing.
        test_transform: The transformation at test time.
        test_target_transform: The target transformation at test time.
        logged_metrics: Metrics logged additional to the default ones.
        seed: See :func:`renate.models.utils.get_generator`.
    """

    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        learning_rate_scheduler: Optional[Optional[Callable[[Optimizer], _LRScheduler]]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        seed: int = defaults.SEED,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ) -> None:
        super().__init__(
            model=model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            batch_size=batch_size,
            logged_metrics=logged_metrics,
            seed=seed,
            mask_unused_classes=mask_unused_classes,
        )
        self._train_transform = train_transform
        self._train_target_transform = train_target_transform
        self._test_transform = test_transform
        self._test_target_transform = test_target_transform
        self._val_memory_buffer: DataBuffer = InfiniteBuffer()

    def _ignored_hyperparameters(self):
        """Hyperparameters to be ignored in the ``save_hyperparameters`` call."""
        return super()._ignored_hyperparameters() + [
            "components",
            "train_transform",
            "train_target_transform",
            "test_transform",
            "test_target_transform",
            "buffer_transform",
            "buffer_target_transform",
        ]

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        learner_state_dict = {
            "learner_class_name": self.__class__.__name__,
            "val_memory_buffer": self._val_memory_buffer.state_dict(),
        }
        checkpoint.update(learner_state_dict)

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]):
        self._val_memory_buffer.load_state_dict(checkpoint["val_memory_buffer"])

    def save(self, output_state_dir: str) -> None:
        val_buffer_dir = os.path.join(output_state_dir, "val_memory_buffer")
        os.makedirs(val_buffer_dir, exist_ok=True)
        self._val_memory_buffer.save(val_buffer_dir)

    def load(self, input_state_dir: str) -> None:
        self._val_memory_buffer.load(os.path.join(input_state_dir, "val_memory_buffer"))

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        super().on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._model.add_task_params(task_id=self._task_id)

    def train_dataloader(self) -> DataLoader:
        """Returns the dataloader for training the model."""
        train_dataset = _TransformedDataset(
            self._train_dataset,
            transform=self._train_transform,
            target_transform=self._train_target_transform,
        )
        return DataLoader(
            train_dataset,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
            collate_fn=self._train_collate_fn,
        )

    def val_dataloader(self) -> Optional[DataLoader]:
        if self._val_dataset is not None:
            val_dataset = _TransformedDataset(
                self._val_dataset,
                transform=self._test_transform,
                target_transform=self._test_target_transform,
            )
            self._val_memory_buffer.update(val_dataset)

        if len(self._val_memory_buffer):
            return DataLoader(
                self._val_memory_buffer,
                batch_size=self._batch_size,
                shuffle=False,
                generator=self._rng,
                pin_memory=True,
                collate_fn=self._val_collate_fn,
            )

    def validation_step_unpack_batch(
        self, batch: Tuple[NestedTensors, torch.Tensor]
    ) -> Tuple[NestedTensors, Any]:
        (inputs, targets), _ = batch
        return inputs, targets


class ReplayLearner(Learner, abc.ABC):
    """Base class for Learners which use a buffer to store data and reuse it in future updates.

    Args:
        memory_size: The maximum size of the memory.
        batch_memory_frac: Fraction of the batch that is sampled from rehearsal memory.
        buffer_transform: The transformation to be applied to the memory buffer data samples.
        buffer_target_transform: The target transformation to be applied to the memory buffer target
            samples.
        seed: See :func:`renate.models.utils.get_generator`.
    """

    def __init__(
        self,
        memory_size: int,
        batch_size: int = defaults.BATCH_SIZE,
        batch_memory_frac: float = defaults.BATCH_MEMORY_FRAC,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        seed: int = defaults.SEED,
        **kwargs,
    ) -> None:
        if not (0 <= batch_memory_frac <= 1):
            raise ValueError(
                f"Expecting batch_memory_frac to be in [0, 1], received {batch_memory_frac}."
            )
        memory_batch_size = min(memory_size, int(batch_memory_frac * batch_size))
        batch_size = batch_size - memory_batch_size
        super().__init__(batch_size=batch_size, seed=seed, **kwargs)
        self._memory_batch_size = memory_batch_size
        self._memory_buffer = ReservoirBuffer(
            max_size=memory_size,
            seed=seed,
            transform=buffer_transform,
            target_transform=buffer_target_transform,
        )

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_save_checkpoint(checkpoint)
        checkpoint["memory_buffer"] = self._memory_buffer.state_dict()

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_load_checkpoint(checkpoint)
        self._memory_buffer.load_state_dict(checkpoint["memory_buffer"])

    def save(self, output_state_dir: str) -> None:
        super().save(output_state_dir)
        buffer_dir = os.path.join(output_state_dir, "memory_buffer")
        os.makedirs(buffer_dir, exist_ok=True)
        self._memory_buffer.save(buffer_dir)

    def load(self, input_state_dir: str) -> None:
        super().load(input_state_dir)
        self._memory_buffer.load(os.path.join(input_state_dir, "memory_buffer"))

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        super().on_model_update_start(
            train_dataset, val_dataset, train_dataset_collate_fn, val_dataset_collate_fn, task_id
        )
        if self._mask_unused_classes:
            self._classes_in_current_task = self._classes_in_current_task.union(
                unique_classes(self._memory_buffer)
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.memory import GreedyClassBalancingBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import ReplayLearner
from renate.updaters.model_updater import SingleTrainingLoopUpdater


class GDumbLearner(ReplayLearner):
    """A Learner that implements the GDumb strategy.

    Prabhu, Ameya, Philip HS Torr, and Puneet K. Dokania. "GDumb: A simple
    approach that questions our progress in continual learning." ECCV, 2020.

    It maintains a memory of  previously observed data points and does
    the training after updating the buffer. Note that, the model is
    reinitialized before training on the buffer.

    Args:
        memory_size: The maximum size of the memory.
        buffer_transform: The transform to be applied to the data points in the memory.
        buffer_target_transform: The transform to be applied to the targets in the memory.
        seed: A random seed.
    """

    def __init__(
        self,
        memory_size: int,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        seed: int = defaults.SEED,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            memory_size=memory_size,
            seed=seed,
            **kwargs,
        )

        self._memory_buffer = GreedyClassBalancingBuffer(
            max_size=memory_size,
            seed=seed,
            transform=buffer_transform,
            target_transform=buffer_target_transform,
        )

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_load_checkpoint(checkpoint)
        self._memory_buffer.load_state_dict(checkpoint["memory_buffer"])

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._memory_buffer.update(train_dataset)

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self._memory_buffer,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
            collate_fn=self._train_collate_fn,
        )

    def training_step(
        self,
        batch: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        batch_idx: int,
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        batch, _ = batch
        return super().training_step(batch=batch, batch_idx=batch_idx)


class GDumbModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=GDumbLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.benchmark.models.l2p import LearningToPromptTransformer, PromptPool, PromptedTransformer


def test_prompt_pool():
    D_emb = 2
    Lp = 3
    feat_dim = 6
    B = 4
    N = 2

    pool = PromptPool(
        embedding_dim=feat_dim, prompt_key_dim=D_emb, prompt_size=Lp, pool_selection_size=N
    )
    input = torch.rand(B, D_emb)
    out = pool(input)[0]

    assert out.shape == (B, Lp * N, feat_dim)


@pytest.mark.slow
def test_prompted_vision_transformer():
    combined = LearningToPromptTransformer()
    inp = torch.rand(1, 3, 224, 224)
    assert combined(inp).shape == torch.Size((1, 10))


@pytest.mark.slow
def test_prompted_text_transformer():
    model = LearningToPromptTransformer(pretrained_model_name_or_path="bert-base-uncased")
    inp = {"input_ids": torch.randint(0, 3000, (10, 128))}
    assert model(inp).shape == torch.Size((10, 10))


@pytest.mark.slow
@pytest.mark.parametrize(
    "cls,arg,argval,error",
    [
        [PromptPool, "similarity_fn", "not_cosine", ValueError],
        [LearningToPromptTransformer, "prompt_embedding_features", "not_cls", AssertionError],
        [LearningToPromptTransformer, "patch_pooler", "not_cls", AssertionError],
    ],
)
def test_pool_vision_transformer_raises_errors(cls, arg, argval, error):
    with pytest.raises(error):
        cls(**{arg: argval})


@pytest.mark.slow
@pytest.mark.parametrize(
    "backbone,num_trainable_params",
    [
        ["google/vit-base-patch16-224", 4],
        ["google/vit-base-patch32-224-in21k", 4],
        ["google/vit-large-patch32-224-in21k", 4],
        ["bert-base-uncased", 4],
        ["distilbert-base-uncased", 4],
    ],
)
def test_prompt_vision_transformer_trainable_parameters(backbone, num_trainable_params):
    # The result is always 4: prompt pool, prompt pool key, classifier wt, classifier bias.
    model = LearningToPromptTransformer(pretrained_model_name_or_path=backbone)
    n = sum(1 for x in model.parameters() if x.requires_grad)
    assert n == num_trainable_params


@pytest.mark.slow
@pytest.mark.parametrize("backbone", ["google/vit-base-patch16-224", "bert-base-uncased"])
@pytest.mark.parametrize("prompt", [None, torch.rand(3, 10, 768)])
@pytest.mark.parametrize("cls_feat", [True, False])
def test_prompted_transformer(backbone, prompt, cls_feat):
    model = PromptedTransformer(
        pretrained_model_name_or_path=backbone,
        num_outputs=10,
        prediction_strategy=None,
        add_icarl_class_means=False,
    )

    B, P_len, _ = prompt.shape if prompt is not None else (5, 0, 0)
    if "vit" in backbone:
        # we are doing ViT.
        inputs = torch.rand(B, 3, 224, 224)
        expected_output_size = (B, 197 + P_len, 768) if not cls_feat else (B, 768)
    else:
        inputs = {"input_ids": torch.randint(0, 10000, (B, 128))}
        expected_output_size = (B, 768)

    out = model(inputs, prompt, cls_feat=cls_feat)
    assert out.shape == expected_output_size



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.shift.kernels import RBFKernel
from renate.shift.mmd_helpers import mmd


def test_mmd_identical_data():
    """We expect high p-values for identical data."""
    X = torch.randn(100, 2)
    _, p_val = mmd(X, X, kernel=RBFKernel(), num_permutations=100)
    assert p_val == 1.0


def test_shift_detector_disjoint_data():
    """We expect low p-values for very different data (two disjoint Gaussian blobs)."""
    X0 = torch.randn(100, 2)
    X1 = torch.randn(100, 2) + 2.0
    _, p_val = mmd(X0, X1, kernel=RBFKernel(), num_permutations=100)
    assert p_val == 0.0


def test_mmd_vs_manual():
    X0 = torch.tensor([[0.0, 0.0], [0.0, 1.0]])
    X1 = torch.tensor([[1.0, 0.0], [2.0, 2.0]])
    mmd_val, _ = mmd(X0, X1, kernel=RBFKernel(lengthscale=1.0), num_permutations=0)
    # Compare against manual computation of the terms in the MMD formula.
    assert mmd_val.item() == pytest.approx(1.6065 + 1.0821 - 2 * 0.2687, abs=1e-4)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import json
import os
from pathlib import Path
import pickle
from typing import List, Literal, Optional, Tuple, Union

import gdown
import pandas as pd
import torch
import torchvision
from torchvision import transforms

from renate import defaults
from renate.benchmark.datasets.base import DataIncrementalDataModule
from renate.data import ImageDataset
from renate.data.data_module import RenateDataModule
from renate.utils.file import (
    download_and_unzip_file,
    download_file,
    download_file_from_s3,
    download_folder_from_s3,
    extract_file,
)


class TinyImageNetDataModule(RenateDataModule):
    """Datamodule that process TinyImageNet dataset.

    Source: http://cs231n.stanford.edu/

    The TinyImageNet dataset is a subset of the ImageNet dataset. It contains 200 classes, each with
    500 training images, 50 validation images with labels. There are also 50 unlabeled test images
    per class, which we are not using here. We use the validation split as the test set.

    Args:
        data_path: Path to the directory where the dataset should be stored.
        src_bucket: Name of the bucket where the dataset is stored.
        src_object_name: Name of the object in the bucket where the dataset is stored.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed to be used for splitting the dataset.
    """

    md5s = {"tiny-imagenet-200.zip": "90528d7ca1a48142e341f4ef8d21d0de"}

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ) -> None:
        super(TinyImageNetDataModule, self).__init__(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = "tiny_imagenet"

    def prepare_data(self) -> None:
        """Download the TinyImageNet dataset."""
        if not self._verify_file("tiny-imagenet-200.zip"):
            download_and_unzip_file(
                self._dataset_name,
                self._data_path,
                self._src_bucket,
                self._src_object_name,
                "http://cs231n.stanford.edu/",
                "tiny-imagenet-200.zip",
            )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        X, y = self._preprocess_tiny_imagenet(train=True)
        train_data = ImageDataset(X, y)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        X, y = self._preprocess_tiny_imagenet(train=False)
        self._test_data = ImageDataset(X, y)

    def _preprocess_tiny_imagenet(self, train: bool) -> Tuple[List[str], List[int]]:
        """A helper function to preprocess the TinyImageNet dataset."""
        if train:
            data_path = os.path.join(self._data_path, self._dataset_name, "tiny-imagenet-200")
        else:
            data_path = os.path.join(self._data_path, self._dataset_name, "tiny-imagenet-200")
        with open(os.path.join(data_path, "wnids.txt"), "r") as f:
            label_encoding = {line.strip(): i for i, line in enumerate(f.readlines())}
        X = []
        y = []
        if train:
            for label in label_encoding.keys():
                class_path = os.path.join(data_path, "train", label)
                for file_name in os.listdir(os.path.join(class_path, "images")):
                    X.append(os.path.join(class_path, "images", file_name))
                    y.append(label_encoding[label])
        else:
            val_annotations_file = os.path.join(data_path, "val", "val_annotations.txt")
            with open(val_annotations_file, "r") as f:
                for line in f:
                    file_name, label = line.split("\t")[:2]
                    X.append(os.path.join(data_path, "val", "images", file_name))
                    y.append(label_encoding[label])
        return X, y


class TorchVisionDataModule(RenateDataModule):
    """Data module wrapping torchvision datasets.

    Args:
        data_path: the path to the folder containing the dataset files.
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        dataset_name: Name of the torchvision dataset.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    dataset_dict = {
        "CIFAR10": (torchvision.datasets.CIFAR10, "cifar-10-batches-py"),
        "CIFAR100": (torchvision.datasets.CIFAR100, "cifar-100-python"),
        "FashionMNIST": (torchvision.datasets.FashionMNIST, "FashionMNIST"),
        "MNIST": (torchvision.datasets.MNIST, "MNIST"),
    }
    dataset_stats = {
        "CIFAR10": {
            "mean": (0.49139967861519607, 0.48215840839460783, 0.44653091444546567),
            "std": (0.24703223246174102, 0.24348512800151828, 0.26158784172803257),
        },
        "CIFAR100": {
            "mean": (0.5070751592371323, 0.48654887331495095, 0.4409178433670343),
            "std": (0.26733428587941854, 0.25643846292120615, 0.2761504713263903),
        },
        "FashionMNIST": {"mean": 0.2860405969887955, "std": 0.3530242445149223},
        "MNIST": {"mean": 0.1306604762738429, "std": 0.30810780385646264},
    }

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        dataset_name: str = "MNIST",
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super(TorchVisionDataModule, self).__init__(
            data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = dataset_name
        if self._dataset_name not in TorchVisionDataModule.dataset_dict:
            raise ValueError(f"Dataset {self._dataset_name} currently not supported.")

    def prepare_data(self) -> None:
        """Download data.

        If s3 bucket is given, the data is downloaded from s3, otherwise from the original source.
        """
        cls, dataset_pathname = TorchVisionDataModule.dataset_dict[self._dataset_name]
        if self._src_bucket is None:
            cls(self._data_path, train=True, download=True)
            cls(self._data_path, train=False, download=True)
        else:
            download_folder_from_s3(
                src_bucket=self._src_bucket,
                src_object_name=self._src_object_name,
                dst_dir=os.path.join(self._data_path, dataset_pathname),
            )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        cls, _ = TorchVisionDataModule.dataset_dict[self._dataset_name]
        train_data = cls(
            self._data_path,
            train=True,
            transform=transforms.ToTensor(),
            target_transform=transforms.Lambda(to_long),
        )
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = cls(
            self._data_path,
            train=False,
            transform=transforms.ToTensor(),
            target_transform=transforms.Lambda(to_long),
        )


def to_long(x):
    return torch.tensor(x, dtype=torch.long)


class CLEARDataModule(DataIncrementalDataModule):
    """Datamodule that process CLEAR datasets: CLEAR10 and CLEAR100.

    Source: https://clear-benchmark.github.io/.

    Args:
        data_path: the path to the folder containing the dataset files.
        time_step: Loads CLEAR dataset for this time step. Options: CLEAR10: [0,9], CLEAR100: [0,10]
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        dataset_name: CLEAR dataset name, options are clear10 and clear100.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    dataset_stats = {
        "CLEAR10": {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]},
        "CLEAR100": {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]},
    }

    md5s = {
        "clear10-train-image-only.zip": "5171f720810d60b471c308dee595d430",
        "clear100-train-image-only.zip": "ea85cdba9efcb3abf77eaab5554052c8",
        "clear10-test.zip": "bf9a85bfb78fe742c7ed32648c9a3275",
        "clear100-test.zip": "e160815fb5fd4bc71dacd339ff41e6a9",
    }

    def __init__(
        self,
        data_path: Union[Path, str],
        time_step: int = 0,
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        dataset_name: str = "CLEAR10",
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super(CLEARDataModule, self).__init__(
            data_path=data_path,
            data_id=time_step,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = dataset_name.lower()
        assert self._dataset_name in ["clear10", "clear100"]
        assert 0 <= self.data_id <= (9 if self._dataset_name == "clear10" else 10)

    def prepare_data(self) -> None:
        """Download CLEAR dataset with given dataset_name (clear10/clear100)."""
        for file_name in [
            f"{self._dataset_name}-train-image-only.zip",
            f"{self._dataset_name}-test.zip",
        ]:
            if not self._verify_file(file_name):
                download_and_unzip_file(
                    self._dataset_name,
                    self._data_path,
                    self._src_bucket,
                    self._src_object_name,
                    "https://huggingface.co/datasets/elvishelvis6/CLEAR-Continual_Learning_Benchmark/resolve/main/",  # noqa: E501
                    file_name,
                )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        time_step = self.data_id + 1 if self._dataset_name == "clear10" else self.data_id
        X, y = self._get_filepaths_and_labels(train=True, time_step=time_step)
        train_data = ImageDataset(X, y)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        X, y = self._get_filepaths_and_labels(train=False, time_step=time_step)
        self._test_data = ImageDataset(X, y)

    def _get_filepaths_and_labels(self, train: bool, time_step: int) -> Tuple[List[str], List[int]]:
        """Extracts all the filepaths and labels for a given chunk id and split."""
        path = os.path.join(self._data_path, self._dataset_name)

        # Load the class names and create a class mapping. The class names are in `class_names.txt`
        with open(os.path.join(path, "train_image_only", "class_names.txt"), "r") as f:
            class_names = [line.strip() for line in f.readlines()]
            label_encoding = {name: cnt for cnt, name in enumerate(class_names)}

        path = os.path.join(path, "train_image_only" if train else "test")
        with open(os.path.join(path, "labeled_metadata.json"), "r") as f:
            metadata = json.load(f)

        image_paths = []
        labels = []
        for class_name, class_metadata_file in metadata[str(time_step)].items():
            label = label_encoding[class_name]
            with open(os.path.join(path, class_metadata_file), "r") as f:
                class_metadata = json.load(f)
            for image_metadata in class_metadata.values():
                image_paths.append(os.path.join(path, image_metadata["IMG_PATH"]))
                labels.append(label)

        return image_paths, labels


class DomainNetDataModule(DataIncrementalDataModule):
    """Datamodule that provides access to DomainNet.

    Args:
        data_path: the path to the folder containing the dataset files.
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        domain: DomainNet domain name, options are clipart, infograph, painting, quickdraw, real,
            and sketch.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    md5s = {
        "clipart.zip": "cd0d8f2d77a4e181449b78ed62bccf1e",
        "clipart_train.txt": "b4349693a7f9c05c53955725c47ed6cb",
        "clipart_test.txt": "f5ddbcfd657a3acf9d0f7da10db22565",
        "infograph.zip": "720380b86f9e6ab4805bb38b6bd135f8",
        "infograph_train.txt": "379b50054f4ac2018dca4f89421b92d9",
        "infograph_test.txt": "779626b50869edffe8ea6941c3755c71",
        "painting.zip": "1ae32cdb4f98fe7ab5eb0a351768abfd",
        "painting_train.txt": "7db0e7ca73ad9982f6e1f7f3ef372c0a",
        "painting_test.txt": "232b35dc53f26d414686ae579e38d9b5",
        "quickdraw.zip": "bdc1b6f09f277da1a263389efe0c7a66",
        "quickdraw_train.txt": "b732ced3939ac8efdd8c0a889dca56cc",
        "quickdraw_test.txt": "c1a828fdfe216fb109f1c0083a252c6f",
        "real.zip": "dcc47055e8935767784b7162e7c7cca6",
        "real_train.txt": "8ebf02c2075fadd564705f0dc7cd6291",
        "real_test.txt": "6098816791c3ebed543c71ffa11b9054",
        "sketch.zip": "658d8009644040ff7ce30bb2e820850f",
        "sketch_train.txt": "1233bd18aa9a8a200bf4cecf1c34ef3e",
        "sketch_test.txt": "d8a222e4672cfd585298aa14d02ea441",
    }

    domains = ["clipart", "infograph", "painting", "quickdraw", "real", "sketch"]
    dataset_stats = {
        "clipart": {"mean": [0.7395, 0.7195, 0.6865], "std": [0.3621, 0.3640, 0.3873]},
        "infograph": {"mean": [0.6882, 0.6962, 0.6644], "std": [0.3328, 0.3095, 0.3277]},
        "painting": {"mean": [0.5737, 0.5456, 0.5067], "std": [0.3079, 0.3003, 0.3161]},
        "quickdraw": {"mean": [0.9525, 0.9525, 0.9525], "std": [0.2127, 0.2127, 0.2127]},
        "real": {"mean": [0.6066, 0.5897, 0.5564], "std": [0.3335, 0.3270, 0.3485]},
        "sketch": {"mean": [0.8325, 0.8269, 0.8180], "std": [0.2723, 0.2747, 0.2801]},
        "all": {"mean": [0.7491, 0.7391, 0.7179], "std": [0.3318, 0.3314, 0.3512]},
    }

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        domain: str = "clipart",
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(
            data_path=data_path,
            data_id=domain.lower(),
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        assert self.data_id in self.domains, f"Unknown domain {self.data_id}."
        self._dataset_name = domain.lower()

    def prepare_data(self) -> None:
        """Download DomainNet dataset for given domain."""
        file_name = f"{self.data_id}.zip"
        # update dataset name:
        self._dataset_name = self.data_id
        url = "http://csr.bu.edu/ftp/visda/2019/multi-source/"
        if self.data_id in ["clipart", "painting"]:
            url = os.path.join(url, "groundtruth")
        if not self._verify_file(file_name):
            download_and_unzip_file(
                self.data_id,
                self._data_path,
                self._src_bucket,
                self._src_object_name,
                url,
                file_name,
            )
        for file_name in [f"{self.data_id}_train.txt", f"{self.data_id}_test.txt"]:
            if not self._verify_file(file_name):
                download_file(
                    self.data_id,
                    self._data_path,
                    self._src_bucket,
                    self._src_object_name,
                    "http://csr.bu.edu/ftp/visda/2019/multi-source/domainnet/txt/",
                    file_name,
                )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        X, y = self._get_filepaths_and_labels("train")
        train_data = ImageDataset(X, y)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        X, y = self._get_filepaths_and_labels("test")
        self._test_data = ImageDataset(X, y)

    def _get_filepaths_and_labels(self, split: str) -> Tuple[List[str], List[int]]:
        """Extracts all the filepaths and labels for a given split."""
        path = os.path.join(self._data_path, self.data_id)
        df = pd.read_csv(
            os.path.join(path, f"{self.data_id}_{split}.txt"),
            sep=" ",
            header=None,
            names=["path", "label"],
        )
        data = list(df.path.apply(lambda x: os.path.join(path, x)))
        labels = list(df.label)
        return data, labels


class CDDBDataModule(DataIncrementalDataModule):
    md5s = {
        "CDDB.tar.zip": "823b6496270ba03019dbd6af60cbcb6b",
    }

    domains = ["gaugan", "biggan", "wild", "whichfaceisreal", "san"]
    dataset_stats = {
        "CDDB": dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    }
    google_drive_id = "1NgB8ytBMFBFwyXJQvdVT_yek1EaaEHrg"

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        domain: str = "gaugan",
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        assert domain in self.domains
        super().__init__(
            data_path=data_path,
            data_id=domain.lower(),
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )

    def prepare_data(self) -> None:
        """Download DomainNet dataset for given domain."""
        file_name = "CDDB.tar.zip"
        self._dataset_name = ""
        if not self._verify_file(file_name):
            if self._src_bucket is None:
                gdown.download(
                    output=self._data_path,
                    quiet=False,
                    url=f"https://drive.google.com/u/0/uc?id={self.google_drive_id}&export=download&confirm=pbef",  # noqa: E501
                )
            else:
                download_file_from_s3(
                    dst=os.path.join(self._data_path, file_name),
                    src_bucket=self._src_bucket,
                    src_object_name=self._src_object_name,
                )
            extract_file(data_path=self._data_path, file_name="CDDB.tar.zip", dataset_name="")
            extract_file(data_path=self._data_path, file_name="CDDB.tar", dataset_name="")

    def setup(self) -> None:
        self._dataset_name = "CDDB"  # we need this because zip+tar
        train_path = self._get_filepaths_and_labels("train")
        train_data = torchvision.datasets.ImageFolder(train_path)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        test_path = self._get_filepaths_and_labels("val")
        self._test_data = torchvision.datasets.ImageFolder(test_path)

    def _get_filepaths_and_labels(self, split: str) -> Tuple[List[str], List[int]]:
        return os.path.join(self._data_path, self._dataset_name, self.data_id, split)


class CORE50DataModule(DataIncrementalDataModule):
    """Datamodule that process the CORe50 dataset.

    It enables to download all the scenarios and with respect to 0th run (as per S-Prompts),
    set by `scenario` and `data_id` respectively.

    Source: https://vlomonaco.github.io/core50/.
    Adapted from: https://github.com/vlomonaco/core50/blob/master/scripts/python/data_loader.py

    Args:
        data_path: The path to the folder containing the dataset files.
        src_bucket: The name of the s3 bucket. If not provided, downloads the data from
            original source.
        src_object_name: The folder path in the s3 bucket.
        scenario: One of ``ni``, ``nc``, ``nic``, ``nicv2_79``, ``nicv2_196`` and ``nicv2_391``.
            This is different from the usage of scenario elsewhere in Renate.
        data_id: One of the several data batches dependent on scenario
    """

    md5s = {
        "core50_128x128.zip": "745f3373fed08d69343f1058ee559e13",
        "paths.pkl": "b568f86998849184df3ec3465290f1b0",
        "LUP.pkl": "33afc26faa460aca98739137fdfa606e",
        "labels.pkl": "281c95774306a2196f4505f22fd60ab1",
    }
    dataset_stats = {
        "Core50": dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    }

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        scenario: Literal["ni", "nc", "nic", "nicv2_79", "nicv2_196", "nicv2_391"] = "ni",
        data_id: int = 0,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ) -> None:
        super().__init__(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            data_id=data_id,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = "core50"
        self._image_source = "core50_128x128"
        self._scenario = scenario
        self._complete_data_path = os.path.join(
            self._data_path, self._dataset_name, self._image_source
        )

    def prepare_data(self) -> None:
        """Download the CORE50 dataset and supporting files and set paths."""
        if not self._verify_file(f"{self._image_source}.zip"):
            download_and_unzip_file(
                self._dataset_name,
                self._data_path,
                self._src_bucket,
                self._src_object_name,
                "http://bias.csr.unibo.it/maltoni/download/core50/",
                f"{self._image_source}.zip",
            )
        for file_name in [
            "paths.pkl",
            "LUP.pkl",
            "labels.pkl",
        ]:
            if not self._verify_file(file_name):
                download_file(
                    self._dataset_name,
                    self._data_path,
                    self._src_bucket,
                    self._src_object_name,
                    "https://vlomonaco.github.io/core50/data/",
                    file_name,
                )

    def setup(self) -> None:
        """Make assignments: train/test splits (CORe50 dataset only has train and test splits)."""
        with open(os.path.join(self._data_path, self._dataset_name, "paths.pkl"), "rb") as f:
            self._paths = pickle.load(f)

        with open(os.path.join(self._data_path, self._dataset_name, "LUP.pkl"), "rb") as f:
            self._LUP = pickle.load(f)

        with open(os.path.join(self._data_path, self._dataset_name, "labels.pkl"), "rb") as f:
            self._labels = pickle.load(f)

        train_data = ImageDataset(*self._parse_file_lists("train"))
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = ImageDataset(*self._parse_file_lists("test"))

    def _parse_file_lists(self, stage: str) -> Tuple[List[str], List[int]]:
        data_id = self.data_id if stage == "train" else -1
        idx_list = self._LUP[self._scenario][0][data_id]
        X = [os.path.join(self._complete_data_path, self._paths[idx]) for idx in idx_list]
        y = self._labels[self._scenario][0][data_id]

        return X, y



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import List, Optional, Type, Union

import torch.nn as nn
from torchvision.models.resnet import BasicBlock, Bottleneck
from torchvision.models.resnet import ResNet as _ResNet

from renate.benchmark.models.base import RenateBenchmarkingModule
from renate.models.prediction_strategies import PredictionStrategy


class ResNet(RenateBenchmarkingModule):
    """ResNet model base class.

    TODO: Fix citation
    Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun:
    Deep Residual Learning for Image Recognition. CVPR 2016: 770-778

    Args:
        block: The type of the block to use as the core building block.
        layers: The number of blocks in the respective parts of ResNet.
        num_outputs: The number of output units.
        zero_init_residual: Whether to set the initial weights of the residual blocks to zero
            through initializing the Batch Normalization parameters at the end of the block to zero.
        groups: The number of groups to be used for the group convolution.
        width_per_group: The width of the group convolution.
        replace_stride_with_dilation: Whether to replace the stride with a dilation to save memory.
        norm_layer: What kind of normalization layer to use, following convolutions.
        cifar_stem: Whether to use a stem for CIFAR-sized images.
        gray_scale: Whether input images are gray-scale images, i.e. only 1 color channel.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
    """

    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        num_outputs: int = 10,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Type[nn.Module] = nn.BatchNorm2d,
        cifar_stem: bool = True,
        gray_scale: bool = False,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ) -> None:
        model = _ResNet(
            block=block,
            layers=layers,
            num_classes=num_outputs,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            replace_stride_with_dilation=replace_stride_with_dilation,
            norm_layer=norm_layer,
        )
        super().__init__(
            embedding_size=model.fc.in_features,
            num_outputs=num_outputs,
            constructor_arguments={
                "block": block,
                "layers": layers,
                "zero_init_residual": zero_init_residual,
                "groups": groups,
                "width_per_group": width_per_group,
                "replace_stride_with_dilation": replace_stride_with_dilation,
                "norm_layer": norm_layer,
                "cifar_stem": cifar_stem,
                "gray_scale": gray_scale,
            },
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )
        self._backbone = model
        if cifar_stem:
            self._backbone.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
            self._backbone.maxpool = nn.Identity()
        if gray_scale:
            self._backbone.conv1 = nn.Conv2d(
                1,
                self._backbone.conv1.out_channels,
                kernel_size=self._backbone.conv1.kernel_size,
                stride=self._backbone.conv1.stride,
                padding=self._backbone.conv1.padding,
                bias=self._backbone.conv1.bias is not None,
            )
        self._backbone.fc = nn.Identity()

        for m in self.modules():
            if hasattr(m, "reset_parameters"):
                m.reset_parameters()


class ResNet18CIFAR(ResNet):
    def __init__(self, block=BasicBlock, layers=[2, 2, 2, 2], cifar_stem=True, **kwargs) -> None:
        super().__init__(block=block, layers=layers, cifar_stem=cifar_stem, **kwargs)


class ResNet34CIFAR(ResNet):
    def __init__(self, block=BasicBlock, layers=[3, 4, 6, 3], cifar_stem=True, **kwargs) -> None:
        super().__init__(block=block, layers=layers, cifar_stem=cifar_stem, **kwargs)


class ResNet50CIFAR(ResNet):
    def __init__(self, block=Bottleneck, layers=[3, 4, 6, 3], cifar_stem=True, **kwargs) -> None:
        super().__init__(block=block, layers=layers, cifar_stem=cifar_stem, **kwargs)


class ResNet18(ResNet):
    def __init__(self, block=BasicBlock, layers=[2, 2, 2, 2], cifar_stem=False, **kwargs) -> None:
        super().__init__(block=block, layers=layers, cifar_stem=cifar_stem, **kwargs)


class ResNet34(ResNet):
    def __init__(self, block=BasicBlock, layers=[3, 4, 6, 3], cifar_stem=False, **kwargs) -> None:
        super().__init__(block=block, layers=layers, cifar_stem=cifar_stem, **kwargs)


class ResNet50(ResNet):
    def __init__(self, block=Bottleneck, layers=[3, 4, 6, 3], cifar_stem=False, **kwargs) -> None:
        super().__init__(block=block, layers=layers, cifar_stem=cifar_stem, **kwargs)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.benchmark.models.transformer import HuggingFaceSequenceClassificationTransformer


@pytest.mark.slow
@pytest.mark.parametrize("model_name", ["distilbert-base-uncased", "bert-base-uncased"])
def test_init(model_name):
    HuggingFaceSequenceClassificationTransformer(
        pretrained_model_name_or_path=model_name, num_outputs=10
    )


@pytest.mark.slow
@pytest.mark.parametrize(
    "model_name,input_dim",
    [
        ["distilbert-base-uncased", (128,)],
        ["bert-base-uncased", (256,)],
    ],
)
def test_text_transformer_fwd(model_name, input_dim):
    transformer = HuggingFaceSequenceClassificationTransformer(
        pretrained_model_name_or_path=model_name
    )

    x = {"input_ids": torch.randint(0, 30000, (5, *input_dim))}
    y_hat = transformer(x)

    assert y_hat.shape[0] == 5
    assert y_hat.shape[1] == 10



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial

import torch
import transformers
from torch.optim import Adam

from renate.benchmark.datasets.nlp_datasets import HuggingFaceTextDataModule
from renate.models.renate_module import RenateWrapper
from renate.updaters.experimental.fine_tuning import FineTuningModelUpdater

# Create model.
transformer_model = transformers.DistilBertForSequenceClassification.from_pretrained(
    "distilbert-base-uncased", num_labels=2, return_dict=False
)
model = RenateWrapper(transformer_model)
loss_fn = torch.nn.CrossEntropyLoss(reduction="none")

# Prepare data.
tokenizer = transformers.DistilBertTokenizer.from_pretrained("distilbert-base-uncased")
data_module = HuggingFaceTextDataModule(
    "data", dataset_name="rotten_tomatoes", tokenizer=tokenizer, val_size=0.2
)
data_module.prepare_data()  # For multi-GPU, call only on rank 0.
data_module.setup()

# Instantiate renate ModelUpdater and run fine-tuning.
optimizer = partial(Adam, learning_rate=3e-4)
updater = FineTuningModelUpdater(
    model,
    loss_fn,
    optimizer=optimizer,
    batch_size=32,
    max_epochs=3,
    input_state_folder=None,
    output_state_folder="renate_output",
)
updater.update(data_module.train_data(), data_module.val_data())


# Do something with model, e.g., save its weights for later use.
torch.save(model.state_dict(), "model_weights.pt")
